/* ClassInfo.java
 * Part of BIT -- Bytecode Instrumenting Tool
 *
 * Copyright (c) 1997, The Regents of the University of Colorado. All
 * Rights Reserved.
 * 
 * Permission to use and copy this software and its documentation for
 * NON-COMMERCIAL purposes and without fee is hereby granted provided
 * that this copyright notice appears in all copies. If you wish to use
 * or wish to have others use BIT for commercial purposes please contact,
 * Stephen V. O'Neil, Director, Office of Technology Transfer at the
 * University of Colorado at Boulder (303) 492-5647.
 *  
 * By downloading BIT, the User agrees and acknowledges that in no event
 * will the Regents of the University of Colorado be liable for any
 * damages including lost profits, lost savings or other indirect,
 * incidental, special or consequential damages arising out of the use or
 * inability to use the BIT software.
 * 
 * BIT was invented by Han Bok Lee at the University of Colorado in
 * Boulder, Colorado.
 */

package BIT.highBIT;

import BIT.lowBIT.*;
import java.util.*;

/**
 * Contains information about a Java <tt>.class</tt> file.
 * 
 * @author  <a href="mailto:hanlee@cs.colorado.edu">Han B. Lee</a>
 * @see classFile.ClassFile
 * @see Routine
 **/

public class ClassInfo {
    /** 
     * Contains parsed class structure.
     *
     * @see classFile.ClassFile
     */
    protected ClassFile classfile;
    /**
     * Name of this class.
     * <br>
     * This value can be read by invoking getClassName method.
     * @see getClassName
     */
    protected String class_name;
    /**
     * Name of super class.
     * <br>
     * This value can be read by invoking getSuperClassName method.
     * @see getSuperClassName
     */
    protected String superclass_name;
    /**
     * Name of the source file.
     * <br>
     * This value can be read by invoking getSourceFileName method.
     * @see getSourceFileName
     */
    protected String source_file_name;
    /**
     * Vector of methods defined in this class file.
     * <br>
     * This value is set when the constructor is invoked and can be
     * read by invoking getRoutines method.
     * @see getRoutines
     * @see java.util.Vector
     */
    protected Vector routines;

    /**
     * Creates a new ClassInfo class which breaks down a <tt>.class</tt>
     * into more easily manageble entities such as Routines.
     * <br>
     * The only attribute that it looks at is the "SourceFile" attribute.
     *
     * @param	filename	the filename of the <tt>.class</tt> file
     *						to be analyzed
     * @see classFile.ClassFile
     * @see classFile.CONSTANT_Utf8_Info
     * @see classFile.CONSTANT_Class_Info
     * @see Routine
     * @see java.util.Vector
     */
    public ClassInfo(String filename) {
        // create an instance of ClassFile
        //   classfile now contains parsed class structure
        classfile = new ClassFile(filename);
        

        CONSTANT_Class_Info class_info = null; ///////

        // get class name
//        CONSTANT_Class_Info class_info = (CONSTANT_Class_Info) classfile.constant_pool[classfile.this_class];
        
        try{
          // get class name
          class_info = (CONSTANT_Class_Info) classfile.constant_pool[classfile.this_class];
        } catch(Exception e) {
          System.err.println("Exception2! in ClassInfo constructor");
          System.err.println(e.getMessage());
          System.out.println(e.toString());
          e.printStackTrace();
        }

        CONSTANT_Utf8_Info info = (CONSTANT_Utf8_Info) classfile.constant_pool[class_info.name_index];
        class_name = new String(info.bytes, 0);

        // get super class name
        class_info = (CONSTANT_Class_Info) classfile.constant_pool[classfile.super_class];
        info = (CONSTANT_Utf8_Info) classfile.constant_pool[class_info.name_index];
        superclass_name = new String(info.bytes, 0);
        
        // look for source file attribute and set source_file_name
        //   accordingly
        for (int i = 0; i < classfile.attributes_count; i++) {
            Attribute_Info attribute = classfile.attributes[i];
            info = (CONSTANT_Utf8_Info) classfile.constant_pool[attribute.attribute_name_index];
            String attribute_name = new String(info.bytes, 0);
            if (attribute_name.equals("SourceFile")) {
                SourceFile_Attribute sourcefile_attribute = (SourceFile_Attribute) attribute;
                info = (CONSTANT_Utf8_Info) classfile.constant_pool[sourcefile_attribute.sourcefile_index];
                source_file_name = new String(info.bytes, 0);
            }
        }
        
        // get the method information and put them in the routines vector
        routines = new Vector();
        for (int i = 0; i < classfile.methods_count; i++) {
            routines.addElement(new Routine(classfile.methods[i], classfile.constant_pool, this));
        }
        
    }
    
    /**
     * Returns the name of this class.
     */
    public String getClassName() {
        return class_name;
    }
    
    /**
     * Returns the name of the super class.
     */
    public String getSuperClassName() {
        return superclass_name;
    }
    
    /**
     * Returns the name of the source file.
     */
    public String getSourceFileName() {
        return source_file_name;
    }
    
    /** 
     * Returns the routines in this class.
     *
     * @see java.util.Vector 
     */
    public Vector getRoutines() {
        return routines;
    }
    
    /**
     * Returns the number of routiens in this class.
     */
    public int getRoutineCount() {
        return routines.size();
    }

    public Cp_Info[] getConstantPool() {
        return classfile.constant_pool;
    }

    /**
     * Adds an entry to the constant pool.
     *
     * @param   the constant pool entry to be addded
     * @return  the new array where the new entry has been added
     * @see BIT.lowBIT.Cp_Info
     */
    public Cp_Info[] addConstantPoolEntry(Cp_Info entry) {
        Cp_Info[] cps = classfile.addConstantPoolEntry(entry);
        for (Enumeration e = routines.elements(); e.hasMoreElements(); ) {
            Routine r = (Routine) e.nextElement();
            r.setConstantPool(cps);
        }
        return cps;
    }
    
    /**
     * Outputs the class file structure to a file.
     *
     * @param   the name of the output file
     */
    public void write(String filename) {
        for (Enumeration e = routines.elements(); e.hasMoreElements(); ) {
            Routine r = (Routine) e.nextElement();
            if (r.getCodeLength() != 0) {
                r.writeReady();
            }
        }
        classfile.write(filename);
    }
    
    /**
     * Allows one to add a method before this class.
     *
     * @param       String representing the name of the class of the method to be added
     * @param       String the name of the method itself
     * @param       Object to be passed to the method as argument
     */
    public void addBefore(String classname, String methodname, Object arg) {
        for (Enumeration e = routines.elements(); e.hasMoreElements(); ) {
            Routine r = (Routine) e.nextElement();
            if (r.getMethodName().equals("<clinit>")) {
                r.addBefore(classname, methodname, arg);
                return;
            }

        }

        for (Enumeration e = routines.elements(); e.hasMoreElements(); ) {
            Routine r = (Routine) e.nextElement();
            if (r.getMethodName().equals("main")) {
                r.addBefore(classname, methodname, arg);
                return;
            }

        }
        
        for (Enumeration e = routines.elements(); e.hasMoreElements(); ) {
            Routine r = (Routine) e.nextElement();
            if (r.getMethodName().equals("<init>")) {
                r.addBefore(classname, methodname, arg);
                return;
            }
        }
    }
    
    /**
     * Allows one to add a method after this class.
     *
     * <B>Important!</B>
     * Because a Java program typically consists of different class files,
     * the meaning of adding a call to a method after a class is not clear.
     * Here we define adding a method after a class to mean that a call 
     * to a method will be inserted before end of this class only if this class
     * contains the <B>main</B> method.
     * In particular, one should note that if a program exits through a call
     * to System.exit(), Runtime.exit(), or through other abnormal ways,
     * it cannot be guaranteed that the call inserted will get executed at all
     * before program termination.
     * One workaround is to instrument System.exit() and Runtime.exe() in your
     * own standard Java class library.
     * Please email us if you know a good way to fix this semantic problem.
     *
     * @param       String representing the name of the class of the method to be added
     * @param       String the name of the method itself
     * @param       Object to be passed to the method as argument
     */
    public void addAfter(String classname, String methodname, Object arg) {
        for (Enumeration e = routines.elements(); e.hasMoreElements(); ) {
            Routine r = (Routine) e.nextElement();
            if (r.getMethodName().equals("main")) {
                r.addAfter(classname, methodname, arg);
            }
            for (Enumeration i = (r.getInstructionArray()).elements(); i.hasMoreElements(); ) {
                Instruction instr = (Instruction) i.nextElement();
                if (instr.getOpcode() == 184) {  // invokestatic
                    if (isExit(instr.getOperandValue())) {
                        instr.addBefore(classname, methodname, arg);
                    }
                }
            }
        }
    }    
    
    public boolean isExit(int index) {
        CONSTANT_Methodref_Info method_info = (CONSTANT_Methodref_Info) classfile.constant_pool[index];
        CONSTANT_Class_Info class_info = 
            (CONSTANT_Class_Info) classfile.constant_pool[method_info.class_index];
        CONSTANT_NameAndType_Info name_type_info =
            (CONSTANT_NameAndType_Info) classfile.constant_pool[method_info.name_and_type_index];
        CONSTANT_Utf8_Info class_utf8_info = 
            (CONSTANT_Utf8_Info) classfile.constant_pool[class_info.name_index];
        CONSTANT_Utf8_Info method_utf8_info = 
            (CONSTANT_Utf8_Info) classfile.constant_pool[name_type_info.name_index];

        String class_name = new String(class_utf8_info.bytes, 0);
        String method_name = new String(method_utf8_info.bytes, 0);
        
        if (method_name.equals("exit") && 
            (class_name.equals("java/lang/System") || class_name.equals("java/lang/Runtime"))) {
            return true;
        }
        return false;
    }

    // Chandra's additions
/*    public int size() {
      return classfile.size();
    }

    public int method_size() {
      return classfile.method_size();
    }

    public int code_length() {
      return classfile.code_length();
    }
    
    public ClassFile getClassFile() {
      return classfile;
    }
    */
}
