/* ClassFile.java
 * Part of BIT -- Bytecode Instrumenting Tool
 *
 * Copyright (c) 1997, The Regents of the University of Colorado. All
 * Rights Reserved.
 * 
 * Permission to use and copy this software and its documentation for
 * NON-COMMERCIAL purposes and without fee is hereby granted provided
 * that this copyright notice appears in all copies. If you wish to use
 * or wish to have others use BIT for commercial purposes please contact,
 * Stephen V. O'Neil, Director, Office of Technology Transfer at the
 * University of Colorado at Boulder (303) 492-5647.
 *  
 * By downloading BIT, the User agrees and acknowledges that in no event
 * will the Regents of the University of Colorado be liable for any
 * damages including lost profits, lost savings or other indirect,
 * incidental, special or consequential damages arising out of the use or
 * inability to use the BIT software.
 * 
 * BIT was invented by Han Bok Lee at the University of Colorado in
 * Boulder, Colorado.
 */

package BIT.lowBIT;

import java.io.*;
import java.util.*;

public class ClassFile {
  //
  // data members as described on page 84 of
  // "The Java Virtual Machine Specification"
  // by Tim Lindhom and Frank Yellin
  //

  /**
   * 0xCAFEBABE identifies the class file format. 
   */
  public int magic;
  /** 
   * Minor version of the compiler used to create this class file. 
   */
  public short minor_version;
  /** 
   * Major version of the compiler used to create this class file. 
   */
  public short major_version;
  /** 
   * The number of entries in the contant pool table.
   * <br>
   * @see ClassFile#constant_pool
   */
  public short constant_pool_count;
  /** 
   * This is where class file keeps its data such as names of methods,
   * fields, class, and other constants.
   */
  public Cp_Info constant_pool[];
  /** 
   * Access modifier for this class.
   */
  public short access_flags;
  /**
   * Index into the constant pool containing information about current class.
   * <br>
   * @see ClassFile#constant_pool
   */
  public short this_class;
  /**
   * Index into the constant pool containing information about this class's super class.
   * <br>
   * @see ClassFile#constant_pool
   */
  public short super_class;
  /**
   * Represents the number of superinterfaces of this class.
   */
  public short interface_count;
  /**
   * Each entry is an index into the constant pool table and contains information
   * about the interface.
   * <br>
   * @see ClassFile#constant_pool
   */
  public short interfaces[];
  /**
   * Represents the number of fields present in this class.
   */
  public short field_count;
  /**
   * Each entry is a field_info structure and contains information
   * about each field in this class.
   * <br>
   * @see Field_Info
   */
  public Field_Info fields[];
  /**
   * Represents the number of methods present in this class.
   */
  public short methods_count;
  /**
   * Each entry is a method_info structure and contains information
   * about each method in this class.
   * <br>
   * @see Method_Info
   */
  public Method_Info methods[];
  /**
   * Represents the number of attributes present in this class.
   */
  public short attributes_count;
  /**
   * Each entry is an attribute_info structure and contains information
   * about each attribute in this class.
   */
  public Attribute_Info attributes[];
  
  /**
   * Constructor reads in a class file and parses it into its internal representation.
   *
   * @param     the name of the class file to be parsed
   */
  public ClassFile(String filename) {
    try {
      DataInputStream iStream = new DataInputStream(new FileInputStream(filename));
      // start reading info from the input stream
      // magic number
      try {
        magic = iStream.readInt();
        if (magic != (int) 0xcafebabe)
          throw new ClassFileException("Invalid .class file");
      }
      catch (ClassFileException e) {
        System.out.println(e.getMessage());
      }
      
      // version number of the compiler that produced this class file
      minor_version = (short) iStream.readUnsignedShort();
      major_version = (short) iStream.readUnsignedShort();
      
      // constant pool
      constant_pool_count = (short) iStream.readUnsignedShort();
      constant_pool = new Cp_Info[constant_pool_count];
      
      // parse constant pool
      Constant_Pool_Parse cp_parse = new Constant_Pool_Parse(constant_pool, iStream);
      
      // access flags
      access_flags = (short) iStream.readUnsignedShort();
      
      // classes
      this_class = (short) iStream.readUnsignedShort();
      super_class = (short) iStream.readUnsignedShort();
      
      // interfaces
      interface_count = (short) iStream.readUnsignedShort();
      interfaces = new short[interface_count];
      
      for (int i = 0; i < interface_count; i++) {
        interfaces[i] = (short) iStream.readUnsignedShort();
      }
      
      // fields
      field_count = (short) iStream.readUnsignedShort();
      fields = new Field_Info[field_count];
      
      for (int i = 0; i < field_count; i++) {
        fields[i] = new Field_Info(constant_pool, iStream);
      }
      
      // methods
      methods_count = (short) iStream.readUnsignedShort();
      methods = new Method_Info[methods_count];
      
      for (int i = 0; i < methods_count; i++) {
        methods[i] = new Method_Info(constant_pool, iStream);
      }
      
      // attributes
      attributes_count = (short) iStream.readUnsignedShort();
      attributes = new Attribute_Info[attributes_count];
      
      Attribute_Info_Parse ai_parse = new Attribute_Info_Parse(attributes, constant_pool, iStream);
    }
    catch (EOFException e) {
      System.out.println("Unexpected End of File (EOF)");
      System.out.println(e.getMessage());
    }
    catch (FileNotFoundException e) {
      System.out.println("File not found");
      System.out.println(e.getMessage());
    }
    catch (IOException e) {
      System.out.println("Other IO exceptions");
      System.out.println(e.getMessage());
    }
  }

  /**
   * Outputs the internal representation of this ClassFile to a file.
   *
   * @param     the name of the file to be created
   */
  public void write(String filename)
  {
    try {
      DataOutputStream oStream = new DataOutputStream(new FileOutputStream(filename));
      // start writing info to the output stream
      // magic number
      oStream.writeInt(magic);
      
      // version number of the compiler that produced this class file
      oStream.writeShort((int) minor_version);
      oStream.writeShort((int) major_version);
      
      // constant pool
      oStream.writeShort((int) constant_pool_count);
      for (int i = 1; i < constant_pool_count; i++) {
        switch (constant_pool[i].getTag()) {
        case Constant.CONSTANT_Class:
          ((CONSTANT_Class_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_Fieldref:
          ((CONSTANT_Fieldref_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_Methodref:
          ((CONSTANT_Methodref_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_InterfaceMethodref:
          ((CONSTANT_InterfaceMethodref_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_String:
          ((CONSTANT_String_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_Integer:
          ((CONSTANT_Integer_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_Float:
          ((CONSTANT_Float_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_Long:
          ((CONSTANT_Long_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_Double:
          ((CONSTANT_Double_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_NameAndType:
          ((CONSTANT_NameAndType_Info) constant_pool[i]).write(oStream);
          break;
        case Constant.CONSTANT_Utf8:
          ((CONSTANT_Utf8_Info) constant_pool[i]).write(oStream);
          break;
        default:
          break;
        }
      }

      // access flags
      oStream.writeShort((int) access_flags);

      // classes
      oStream.writeShort((int) this_class);
      oStream.writeShort((int) super_class);

      // interfaces
      oStream.writeShort((int) interface_count);
      for (int i = 0; i < interface_count; i++) {
        oStream.writeShort((int) interfaces[i]);
      }

      // fields
      oStream.writeShort((int) field_count);
      for (int i = 0; i < field_count; i++) {
        fields[i].write(constant_pool, oStream);
      }
      
      // methods
      oStream.writeShort((int) methods_count);
      for (int i = 0; i < methods_count; i++) {
        methods[i].write(constant_pool, oStream);
      }

      // attributes
      oStream.writeShort((int) attributes_count);
      for (int i = 0; i < attributes_count; i++) {
        CONSTANT_Utf8_Info attribute_name = (CONSTANT_Utf8_Info)
          constant_pool[attributes[i].attribute_name_index];
        String attribute_string = new String(attribute_name.bytes, 0);
        // since switch does not work on String object
        if (attribute_string.equals("SourceFile")) {
          ((SourceFile_Attribute) attributes[i]).write(oStream);
        }
        else if (attribute_string.equals("ConstantValue")) {
          ((ConstantValue_Attribute) attributes[i]).write(oStream);
        }
        else if (attribute_string.equals("Code")) {
          ((Code_Attribute) attributes[i]).write(constant_pool, oStream);
        }
        else if (attribute_string.equals("Exceptions")) {
          ((Exceptions_Attribute) attributes[i]).write(oStream);
        }
        else if (attribute_string.equals("LineNumberTable")) {
          ((LineNumberTable_Attribute) attributes[i]).write(oStream);
        }
        else if (attribute_string.equals("LocalVariableTable")) {
          ((LocalVariableTable_Attribute) attributes[i]).write(oStream);
        }
        else {			// if unrecognizable attribute, just ignore them
          ((Unknown_Attribute) attributes[i]).write(oStream);
        }
      }
    }
    catch (IOException e) {
      System.out.println("IO exceptions");
      System.out.println(e.getMessage());
    }
  }
  
  /**
   * Allows one to add a contant pool entry (Cp_Info) to the constant pool table.
   *
   * @param     the constant pool entry to be added to the table
   * @return    the new constant pool table
   * @see       Cp_Info
   */
  public Cp_Info[] addConstantPoolEntry(Cp_Info entry) {
    Cp_Info new_cp[];
    
    new_cp = new Cp_Info[constant_pool_count];
    // copy constant pool array into new_cp
    System.arraycopy(constant_pool, 1, new_cp, 1, constant_pool_count - 1);

    constant_pool = new Cp_Info[++constant_pool_count];
    System.arraycopy(new_cp, 1, constant_pool, 1, constant_pool_count - 2);

    constant_pool[constant_pool_count-1] = entry;
    return constant_pool;
  }
}