/* Code_Attribute.java
 * Part of BIT -- Bytecode Instrumenting Tool
 *
 * Copyright (c) 1997, The Regents of the University of Colorado. All
 * Rights Reserved.
 * 
 * Permission to use and copy this software and its documentation for
 * NON-COMMERCIAL purposes and without fee is hereby granted provided
 * that this copyright notice appears in all copies. If you wish to use
 * or wish to have others use BIT for commercial purposes please contact,
 * Stephen V. O'Neil, Director, Office of Technology Transfer at the
 * University of Colorado at Boulder (303) 492-5647.
 *  
 * By downloading BIT, the User agrees and acknowledges that in no event
 * will the Regents of the University of Colorado be liable for any
 * damages including lost profits, lost savings or other indirect,
 * incidental, special or consequential damages arising out of the use or
 * inability to use the BIT software.
 * 
 * BIT was invented by Han Bok Lee at the University of Colorado in
 * Boulder, Colorado.
 */

package BIT.lowBIT;

import java.io.*;

/**
 * One of the attributes that represents the code inside a method.
 * This attribute contains the actual code buffer as well as information
 * about exception tables.
 * <br>
 * @see Attribute_Info
 */
public class Code_Attribute extends Attribute_Info {
  /**
   * max_stack indicates the maximum number of words on the operand
   * statck at any point during execution of this method.
   */
  public short max_stack;
  /** 
   * max_locals indicates the number of local vairables used by this method.
   */
  public short max_locals;
  /**  
   * code-length indicates the number of bytes in the code array for this 
   * method.
   */
  public int code_length;
  /**
   * code array is the code (what else? :).
   */
  public byte code[];
  /**
   * exception_table_length indicates the number of entries in the
   * exception_table table.
   */
  public short exception_table_length;
  /**
   * exceptions is an array of exception_table.
   *
   * @see   Exception_Table
   */
  public Exception_Table exceptions[];
  /**
   * attribute_count indicates the number of attributes in attributes array.
   */
  public short attribute_count;
  /** 
   * attributes is an array of attribute_info.
   * 
   * @see Attribute_Info
   */
  public Attribute_Info attributes[];
  
  /**
   * Parses this attribute given a DataInputStream.
   *
   * @param     constant pool table for the class
   * @param     DataInputStream to be parsed
   * @param     the index of this attribute into the constant pool
   */
  public Code_Attribute(Cp_Info[] constant_pool, DataInputStream iStream,
    short attribute_name_index) throws IOException {
    this.attribute_name_index = attribute_name_index;
    attribute_length = iStream.readInt();
    max_stack = (short) iStream.readUnsignedShort();
    max_locals = (short) iStream.readUnsignedShort();
    code_length = iStream.readInt();
    code = new byte[code_length];
    iStream.readFully(code);
    exception_table_length = (short) iStream.readUnsignedShort();
    exceptions = new Exception_Table[exception_table_length];
    
    for (int i = 0; i < exception_table_length; i++) {
      exceptions[i] = new Exception_Table(iStream);
    }
    
    attribute_count = (short) iStream.readUnsignedShort();
    attributes = new Attribute_Info[attribute_count];
    
    Attribute_Info_Parse ai_parse = new Attribute_Info_Parse(attributes, constant_pool, iStream);
  }

  /**
   * Given an output stream, it outputs its data to that output stream.
   *
   * @param     constant pool table
   * @param     output stream where to output the data
   */
  public void write(Cp_Info[] constant_pool, DataOutputStream oStream)
    throws IOException {
    oStream.writeShort((int) attribute_name_index);
    oStream.writeInt(attribute_length);
    oStream.writeShort((int) max_stack);
    oStream.writeShort((int) max_locals);
    oStream.writeInt(code_length);
    oStream.write(code, 0, code_length);
    oStream.writeShort((int) exception_table_length);

    for (int i = 0; i < exception_table_length; i++) {
      exceptions[i].write(oStream);
    }

    oStream.writeShort((int) attribute_count);
    for (int i = 0; i < attribute_count; i++) {
      CONSTANT_Utf8_Info attribute_name = (CONSTANT_Utf8_Info)
        constant_pool[attributes[i].attribute_name_index];
      String attribute_string = new String(attribute_name.bytes, 0);
      // since switch does not work on String object
      if (attribute_string.equals("SourceFile")) {
        ((SourceFile_Attribute) attributes[i]).write(oStream);
      }
      else if (attribute_string.equals("ConstantValue")) {
        ((ConstantValue_Attribute) attributes[i]).write(oStream);
      }
      else if (attribute_string.equals("Code")) {
        ((Code_Attribute) attributes[i]).write(constant_pool, oStream);
      }
      else if (attribute_string.equals("Exceptions")) {
        ((Exceptions_Attribute) attributes[i]).write(oStream);
      }
      else if (attribute_string.equals("LineNumberTable")) {
        ((LineNumberTable_Attribute) attributes[i]).write(oStream);
      }
      else if (attribute_string.equals("LocalVariableTable")) {
        ((LocalVariableTable_Attribute) attributes[i]).write(oStream);
      }
      else {			// if unrecognizable attribute, just ignore them
        ((Unknown_Attribute) attributes[i]).write(oStream);
      }
    }
  }
}
