
#include "semantic.h"
#include "table.h"
#include "mp_stacks.h"
#include "mp_ast.h"
#include <strstream.h>

static int method_num = 0;
static FileSymbol *file_symbol;
static wchar_t *class_name;
static wchar_t *proc_name;
static MethodSymbol *proc_symbol;

static StatementStack loop_stack;
static IntStack has_break_stack;

// --------------------------------------------------------------------------

static bool IsFormal(MethodSymbol *method, VariableSymbol *variable)
{
	for (int i=0; i<method->NumFormalParameters(); i++)
		if (method->FormalParameter(i) == variable)
			return true;

	return false;
}

// --------------------------------------------------------------------------

static void MethodOpen(ostream& tex)
{
	tex << "\\MPFile{";
	Unicode::TeXOut(tex, file_symbol -> Name());
	tex << "." << ++method_num << "}%\n";
}

// --------------------------------------------------------------------------

MP_Ast *Ast::MetaPost(LexStream& lex_stream, ostream& tex)
{
    cout << "Node number " << (int) kind << " does not contain a MetaPost routine\n";
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstBlock::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_stats = 0;
	int num_statements = NumStatements();
	MP_Ast *stats[num_statements];

	//    Unicode::Cout(label_token_opt ? lex_stream.Name(label_token_opt) : L"");

    if (num_statements > 0)
    {
		int prev_b;
		for (int k = 0; k < num_statements; k++)
		{
			stats[num_stats] = this -> Statement(k) -> MetaPost(lex_stream, tex);
			if (stats[num_stats])
				num_stats++;
		}
    }
	return new MP_StatSeq(lex_stream.Line(LeftToken()),
						  num_stats, stats);
}

// --------------------------------------------------------------------------

MP_Ast *AstPrimitiveType::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_Type(lex_stream.Line(LeftToken()),
					   lex_stream.Name(primitive_kind_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstArrayType::MetaPost(LexStream& lex_stream, ostream& tex)
{
	if (type -> SimpleNameCast())
		return new MP_Type(lex_stream.Line(LeftToken()),
						   lex_stream.Name(type -> SimpleNameCast() -> identifier_token),
						   this -> NumBrackets());
	else
		return new MP_Type(lex_stream.Line(LeftToken()),
						   lex_stream.Name(type -> PrimitiveTypeCast() -> primitive_kind_token),
						   this -> NumBrackets());
}

// --------------------------------------------------------------------------

MP_Ast *AstSimpleName::MetaPost(LexStream& lex_stream, ostream& tex)
{
	wchar_t *owner_name = NULL;

	if (symbol && symbol -> Kind() == Symbol::VARIABLE)
	{
		owner_name = symbol -> VariableCast() -> owner -> Name();
		if (!wcscmp(owner_name, proc_name))
			if (IsFormal(proc_symbol, symbol -> VariableCast()))
				owner_name = L"\\FormalVar";
			else
				owner_name = L"\\LocalVar";
		else if (!wcscmp(owner_name, class_name))
			owner_name = L"\\MemberVar";
		else if (!wcscmp(owner_name, L"<init>"))
			owner_name = L"\\InitVar";
	}

	return new MP_Identifier(lex_stream.Line(LeftToken()),
							 lex_stream.Name(identifier_token), 
							 Type() ? Type() -> Name() : NULL,
							 owner_name);
}

// --------------------------------------------------------------------------

MP_Ast *AstPackageDeclaration::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstImportDeclaration::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return NULL;
}

// --------------------------------------------------------------------------

void AstCompilationUnit::MetaPost(LexStream& lex_stream, 
								  ostream& mp, 
								  ostream& tex, 
								  bool show_source,
								  FileSymbol *file_sym)
{
	mp << "input gn_util;\n";
	mp << "verbatimtex\n";
	mp << "\\documentclass{article}\n";
	mp << "\\usepackage{mp}\n";
	mp << "\\begin{document}\n";
	mp << "etex;\n\n";

	tex << "\\documentclass{article}\n";
	tex << "\\usepackage{mp}\n";
	tex << "\\begin{document}\n";

	file_symbol = file_sym;

	for (int n = 0; n < this -> NumTypeDeclarations(); n++)
		if (this -> TypeDeclaration(n) -> kind == CLASS)
			this -> TypeDeclaration(n) -> ClassDeclarationCast() -> MetaPost(lex_stream, mp, tex, show_source);
		else
			this -> TypeDeclaration(n) -> MetaPost(lex_stream, tex);

	mp << "\nverbatimtex\n";
	mp << "\\end{document}\n";
	mp << "etex;\n";
	mp << "end\n";

	tex << "\\end{document}\n";
}

// --------------------------------------------------------------------------

MP_Ast *AstModifier::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstEmptyDeclaration::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return NULL;
}

// --------------------------------------------------------------------------

void AstClassBody::MetaPost(LexStream& lex_stream, ostream& mp, ostream& tex, bool show_source)
{
	PrintSemantics(tex, lex_stream);

    for (int k = 0; k < this -> NumClassBodyDeclarations(); k++)
	{
		int kind = this -> ClassBodyDeclaration(k) -> kind;
		if (kind != FIELD && 
			kind != STATIC_INITIALIZER && 
			kind != EMPTY_DECLARATION)
		{
			
			tex << "\\ClassMethod{";
			if (this -> ClassBodyDeclaration(k) -> ConstructorDeclarationCast())
				Unicode::TeXOut(tex, lex_stream.Name(this -> ClassBodyDeclaration(k) -> ConstructorDeclarationCast() -> constructor_declarator -> identifier_token));
			else
				Unicode::TeXOut(tex, lex_stream.Name(this -> ClassBodyDeclaration(k) -> MethodDeclarationCast() -> method_declarator -> identifier_token));
			tex << "}\n";

			this -> ClassBodyDeclaration(k) -> PrintSemantics(tex, lex_stream);
	
			if (show_source)
			{
				tex << "\\begin{Source}\n";
				Ast::verbatim = true;
				this -> ClassBodyDeclaration(k) -> PrettyPrint(tex, lex_stream);
				Ast::verbatim = false;
				tex << "\\end{Source}\n";
			}
			this -> ClassBodyDeclaration(k) -> MetaPost(lex_stream, tex) -> Print(mp);
		}
	}
}

// --------------------------------------------------------------------------

void AstClassDeclaration::MetaPost(LexStream& lex_stream, ostream& mp, ostream& tex, bool show_source)
{
	class_name = lex_stream.Name(identifier_token);
	tex << "\\Class{";
	Unicode::TeXOut(tex, class_name);
    tex << "}\n";

	PrintSemantics(tex, lex_stream);

    class_body -> MetaPost(lex_stream, mp, tex, show_source);
}

// --------------------------------------------------------------------------

MP_Ast *AstArrayInitializer::MetaPost(LexStream& lex_stream, ostream& tex)
{
	ostrstream s;

	this -> PrettyPrint(s, lex_stream);

	char *i = new char[s.pcount() + 1]; // deleted by MP_ArrayInit
	strncpy(i, s.str(), s.pcount());
	i[s.pcount()] = 0;

	return new MP_ArrayInit(lex_stream.Line(LeftToken()),
							i);
}

// --------------------------------------------------------------------------

MP_Ast *AstBrackets::MetaPost(LexStream& lex_stream, ostream& tex)
{
    cout << "[]";
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstVariableDeclaratorId::MetaPost(LexStream& lex_stream, ostream& tex)
{
    return new MP_VariableDeclarationId(lex_stream.Line(LeftToken()),
										lex_stream.Name(identifier_token),
										NumBrackets());
}

// --------------------------------------------------------------------------

MP_Ast *AstVariableDeclarator::MetaPost(LexStream& lex_stream, ostream& tex)
{
    MP_Ast *id = variable_declarator_name -> MetaPost(lex_stream, tex);

    if (variable_initializer_opt)
		return new MP_BinExpr(lex_stream.Line(LeftToken()),
							  id,
							  new MP_Operator(lex_stream.Line(LeftToken()),
											  L"\\leftarrow"),
							  variable_initializer_opt -> MetaPost(lex_stream, tex));
	else
		return id;
}

// --------------------------------------------------------------------------

MP_Ast *AstFieldDeclaration::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstFormalParameter::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_modifiers = this -> NumParameterModifiers();
	wchar_t *modifiers[num_modifiers];
	
    for (int i = 0; i < num_modifiers; i++)
		modifiers[i] = lex_stream.Name(this -> ParameterModifier(i) -> modifier_kind_token);

	MP_Ast *variable_id = variable_declarator_name -> MetaPost(lex_stream, tex);
	return new MP_Formal(lex_stream.Line(LeftToken()),
						 new MP_Modifiers(lex_stream.Line(LeftToken()),
										  num_modifiers, modifiers),
						 new MP_VariableDeclaration(lex_stream.Line(LeftToken()),
													type -> MetaPost(lex_stream, tex),
													1,
													&variable_id));
}

// --------------------------------------------------------------------------

MP_Ast *AstMethodDeclarator::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_formals = this -> NumFormalParameters();
	MP_Ast *formals[num_formals];

    for (int j = 0; j < num_formals; j++)
		formals[j] = this -> FormalParameter(j) -> MetaPost(lex_stream, tex);
	//    for (int i = 0; i < NumBrackets(); i++)
	//		tex << " []";
	return new MP_Formals(lex_stream.Line(LeftToken()),
						  num_formals, formals);
}

// --------------------------------------------------------------------------

MP_Ast *AstMethodDeclaration::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_modifiers = this -> NumMethodModifiers();
	wchar_t *modifiers[num_modifiers];
    for (int i = 0; i < num_modifiers; i++)
		modifiers[i] = lex_stream.Name(this -> MethodModifier(i) -> modifier_kind_token);

	MethodOpen(tex);
	proc_name = lex_stream.Name(method_declarator -> identifier_token);
	proc_symbol = method_symbol;
	return new MP_Procedure(lex_stream.Line(LeftToken()),
							proc_name,
							method_declarator -> MetaPost(lex_stream, tex),
							new MP_Modifiers(lex_stream.Line(LeftToken()),
											 num_modifiers, modifiers),
							type -> MetaPost(lex_stream, tex),
							method_body -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstStaticInitializer::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstThisCall::MetaPost(LexStream& lex_stream, ostream& tex)
{
		MethodSymbol *method_symbol = (MethodSymbol *)symbol;
		int num_args = NumArguments();
		MP_Ast *args[num_args];
		wchar_t *argtypes[num_args];
		int num_throws = method_symbol -> NumThrows();
		wchar_t *throws[num_throws];
		MP_Throws *mp_throws = NULL;

		for (int j = 0; j < num_args; j++)
		{
			args[j] = this -> Argument(j) -> MetaPost(lex_stream, tex);
			argtypes[j] = this -> Argument(j) -> Type() -> Name();
		}

		for (int k = 0; k < num_throws; k++)
			throws[k] = method_symbol -> Throws(k) -> Name();
		if (num_throws)
			mp_throws = new MP_Throws(lex_stream.Line(LeftToken()),
									  num_throws, throws);

		return new MP_ProcCall(lex_stream.Line(LeftToken()),
							   new MP_Identifier(lex_stream.Line(LeftToken()),
												 L"this"),
							   num_args,
							   args,
							   argtypes,
							   NULL,
							   mp_throws);
}

// --------------------------------------------------------------------------

MP_Ast *AstSuperCall::MetaPost(LexStream& lex_stream, ostream& tex)
{
	if (!generated)
	{
		//		if (base_opt)
		//			base_opt -> PrettyPrint(out, lex_stream);
		MethodSymbol *method_symbol = (MethodSymbol *)symbol;
		int num_args = NumArguments();
		MP_Ast *args[num_args];
		wchar_t *argtypes[num_args];
		int num_throws = method_symbol -> NumThrows();
		wchar_t *throws[num_throws];
		MP_Throws *mp_throws = NULL;

		for (int j = 0; j < num_args; j++)
		{
			args[j] = this -> Argument(j) -> MetaPost(lex_stream, tex);
			argtypes[j] = this -> Argument(j) -> Type() -> Name();
		}

		for (int k = 0; k < num_throws; k++)
			throws[k] = method_symbol -> Throws(k) -> Name();
		if (num_throws)
			mp_throws = new MP_Throws(lex_stream.Line(LeftToken()),
									  num_throws, throws);

		return new MP_ProcCall(lex_stream.Line(LeftToken()),
							   new MP_Identifier(lex_stream.Line(LeftToken()),
												 L"super"),
							   num_args,
							   args,
							   argtypes,
							   NULL,
							   mp_throws);
	}

	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstConstructorBlock::MetaPost(LexStream& lex_stream, ostream& tex)
{
    if (explicit_constructor_invocation_opt)
	{
		MP_Ast *stats[2];
		int num_stats = 1;

		stats[0] = explicit_constructor_invocation_opt -> MetaPost(lex_stream, tex);
		if (stats[0])
			num_stats++;
		stats[num_stats - 1] = block -> MetaPost(lex_stream, tex);
		return new MP_StatSeq(lex_stream.Line(LeftToken()),
							  num_stats, stats);
	}
    return block -> MetaPost(lex_stream, tex);
}

// --------------------------------------------------------------------------

MP_Ast *AstConstructorDeclaration::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_modifiers = this -> NumConstructorModifiers();
	wchar_t *modifiers[num_modifiers];
    for (int i = 0; i < num_modifiers; i++)
		modifiers[i] = lex_stream.Name(this -> ConstructorModifier(i) -> modifier_kind_token);

	MethodOpen(tex);
	proc_name = lex_stream.Name(constructor_declarator -> identifier_token);
	proc_symbol = constructor_symbol;
	return new MP_Procedure(lex_stream.Line(LeftToken()),
							proc_name,
							constructor_declarator -> MetaPost(lex_stream, tex),
							new MP_Modifiers(lex_stream.Line(LeftToken()),
											 num_modifiers, modifiers),
							NULL,
							constructor_body -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstInterfaceDeclaration::MetaPost(LexStream& lex_stream, ostream& tex)
{
    for (int i = 0; i < this -> NumInterfaceModifiers(); i++)
    {
		Unicode::Cout(lex_stream.Name(this -> InterfaceModifier(i) -> modifier_kind_token));
		cout << " ";
    }
    Unicode::Cout(lex_stream.Name(interface_token));
    cout << " ";
    Unicode::Cout(lex_stream.Name(identifier_token));
	if (NumExtendsInterfaces() > 0)
	{
		cout << " extends ";
		for (int k = 0; k < NumExtendsInterfaces(); k++)
		{
			this -> ExtendsInterface(k) -> MetaPost(lex_stream, tex);
			if (k < NumExtendsInterfaces() - 1)
				cout << ", ";
		}
	}
    cout << "\n";
	cout << "{\n";
    for (int n = 0; n < NumInterfaceMemberDeclarations(); n++)
		this -> InterfaceMemberDeclaration(n) -> MetaPost(lex_stream, tex);
	cout << "}\n";
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstLocalVariableDeclarationStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_decls = this -> NumVariableDeclarators();
	MP_Ast *decls[num_decls];

    for (int k = 0; k < num_decls; k++)
		decls[k] = this -> VariableDeclarator(k) -> MetaPost(lex_stream, tex);

	return new MP_VariableDeclaration(lex_stream.Line(LeftToken()),
									  type -> MetaPost(lex_stream, tex), 
									  num_decls,
									  decls);
}

// --------------------------------------------------------------------------

MP_Ast *AstIfStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	MP_AstStack expressions, statements;
	AstIfStatement *if_stat = this;
	MP_Ast *false_statement = NULL;

	expressions.Push(expression -> MetaPost(lex_stream, tex));
	statements.Push(true_statement -> MetaPost(lex_stream, tex));
	while (if_stat -> false_statement_opt)
		if (if_stat -> false_statement_opt -> kind == IF ||
			(if_stat -> false_statement_opt -> kind == BLOCK && 
			 if_stat -> false_statement_opt -> BlockCast() -> NumStatements() == 1 && 
			 if_stat -> false_statement_opt -> BlockCast() -> Statement(0) -> kind == IF))
		{
			if (if_stat -> false_statement_opt -> kind == IF)
				if_stat = if_stat -> false_statement_opt -> IfStatementCast();
			else
				if_stat = if_stat -> false_statement_opt -> BlockCast() -> Statement(0) -> IfStatementCast();
			expressions.Push(if_stat -> expression -> MetaPost(lex_stream, tex));
			statements.Push(if_stat -> true_statement -> MetaPost(lex_stream, tex));
		}
		else
		{
			false_statement = if_stat -> false_statement_opt -> MetaPost(lex_stream, tex);
			break;
		}

	int num_ifs = expressions.Size();
	MP_Ast *exprs[num_ifs], *stats[num_ifs];

	for (int i=0; i<num_ifs; i++)
	{
		exprs[i] = expressions[i];
		stats[i] = statements[i];
	}

	return new MP_If(lex_stream.Line(LeftToken()),
					 num_ifs, exprs, stats, false_statement);
}

// --------------------------------------------------------------------------

MP_Ast *AstEmptyStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	//    Unicode::Cout(lex_stream.Name(semicolon_token));
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstExpressionStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
    return expression -> MetaPost(lex_stream, tex);
}

// --------------------------------------------------------------------------

MP_Ast *AstCaseLabel::MetaPost(LexStream& lex_stream, ostream& tex)
{
    return expression -> MetaPost(lex_stream, tex);
}

// --------------------------------------------------------------------------

MP_Ast *AstDefaultLabel::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstSwitchBlockStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_stats = 0;
	int num_statements = NumStatements();
	MP_Ast *stats[num_statements];

	has_break_stack.Push(false);
    if (num_statements > 0)
    {
		int prev_b;
		for (int k = 0; k < num_statements; k++)
		{
			stats[num_stats] = this -> Statement(k) -> MetaPost(lex_stream, tex);
			if (stats[num_stats])
				num_stats++;
		}
		if (Statement(num_statements - 1) -> kind == BREAK)
			num_stats--;
    }
	MP_Ast *body = new MP_StatSeq(lex_stream.Line(LeftToken()),
								  num_stats, stats);
	bool has_break = has_break_stack.Top();
	has_break_stack.Pop();

	// TODO: make work for > 1 labels
	MP_Ast *expression;
    for (int j = 0; j < NumSwitchLabels(); j++)
	{
		expression = this -> SwitchLabel(j) -> MetaPost(lex_stream, tex);
	}

	if (expression)
		return new MP_CaseBranch(lex_stream.Line(LeftToken()),
								 expression, body, has_break);
	else
		return new MP_DefaultCaseBranch(lex_stream.Line(LeftToken()),
										body);
}

// --------------------------------------------------------------------------

MP_Ast *AstSwitchStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_cases = 0;
	int num_statements = switch_block -> NumStatements();
	MP_CaseBranch *cases[num_statements];
	MP_DefaultCaseBranch *default_case = NULL;
	AstSwitchBlockStatement *stat;

	loop_stack.Push(this);
    if (num_statements > 0)
    {
		for (int k = 0; k < num_statements; k++)
		{
			stat = switch_block -> Statement(k) -> SwitchBlockStatementCast();
			cases[num_cases] = (MP_CaseBranch*)stat -> MetaPost(lex_stream, tex);
			if (stat -> SwitchLabel(0) -> kind == DEFAULT)
			{
				default_case = (MP_DefaultCaseBranch*)cases[num_cases];
				break;
			}
			if (cases[num_cases])
				num_cases++;
		}
    }
	loop_stack.Pop();

	return new MP_Case(lex_stream.Line(LeftToken()),
					   expression -> MetaPost(lex_stream, tex),
					   num_cases,
					   cases,
					   default_case);
}

// --------------------------------------------------------------------------

MP_Ast *AstWhileStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	loop_stack.Push(this);
	MP_Ast *stat = statement -> MetaPost(lex_stream, tex);
	loop_stack.Pop();
    return new MP_While(lex_stream.Line(LeftToken()),
						expression -> MetaPost(lex_stream, tex), stat);
						
}

// --------------------------------------------------------------------------

MP_Ast *AstDoStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	loop_stack.Push(this);
	MP_Ast *stat = statement -> MetaPost(lex_stream, tex);
	loop_stack.Pop();
    return new MP_Repeat(lex_stream.Line(LeftToken()),
						 stat, expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstForStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	MP_Ast **inits, **updates, *condition = NULL, *stat;
	int num_inits = NumForInitStatements();
	int num_updates = NumForUpdateStatements();

	if (num_inits)
	{
		inits = new MP_Ast*[num_inits];
		for (int m = 0; m < num_inits; m++)
			inits[m] = this -> ForInitStatement(m) -> MetaPost(lex_stream, tex);
	}
    if (end_expression_opt)
		condition = end_expression_opt -> MetaPost(lex_stream, tex);
	if (num_updates)
	{
		updates = new MP_Ast*[num_updates];
		for (int n = 0; n < num_updates; n++)
			updates[n] = this -> ForUpdateStatement(n) -> MetaPost(lex_stream, tex);
	}
	loop_stack.Push(this);
    stat = statement -> MetaPost(lex_stream, tex);
	loop_stack.Pop();
	return new MP_For(lex_stream.Line(LeftToken()),
					  num_inits, inits, condition, num_updates, updates, stat);
}

// --------------------------------------------------------------------------

MP_Ast *AstBreakStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	if (identifier_token_opt)
	{
	}
	if (loop_stack.Size() && loop_stack.Top() -> kind == SWITCH)
	{
		has_break_stack.Pop();
		has_break_stack.Push(true);
	}

	return new MP_Break(lex_stream.Line(LeftToken()));
}

// --------------------------------------------------------------------------

MP_Ast *AstContinueStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	if (identifier_token_opt)
	{
		//		Unicode::Cout(lex_stream.Name(identifier_token_opt));
	}
	return new MP_Continue(lex_stream.Line(LeftToken()));
}

// --------------------------------------------------------------------------

MP_Ast *AstReturnStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	if (!generated)
	{
		MP_Ast *expr = NULL;

		if (expression_opt)
			expr = expression_opt -> MetaPost(lex_stream, tex);

		return new MP_Return(lex_stream.Line(LeftToken()),
							 expr);
	}
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstThrowStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_Throw(lex_stream.Line(LeftToken()),
						expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstSynchronizedStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
    Unicode::Cout(lex_stream.Name(synchronized_token));
    cout << " (";
    expression -> MetaPost(lex_stream, tex);
	cout << ")";
	block -> MetaPost(lex_stream, tex);
	cout << ";\n";
	return NULL;
}

// --------------------------------------------------------------------------

MP_Ast *AstCatchClause::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_Catch(lex_stream.Line(LeftToken()),
						formal_parameter -> MetaPost(lex_stream, tex),
						block -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstFinallyClause::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return block -> MetaPost(lex_stream, tex);
}

// --------------------------------------------------------------------------

MP_Ast *AstTryStatement::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_catches = this -> NumCatchClauses();
	MP_Catch *catches[num_catches];
	MP_Ast *finally = NULL;

	for (int k = 0; k < num_catches; k++)
		catches[k] = (MP_Catch*)this -> CatchClause(k) -> MetaPost(lex_stream, tex);
    if (finally_clause_opt)
		finally = finally_clause_opt -> MetaPost(lex_stream, tex);

	return new MP_Try(lex_stream.Line(LeftToken()),
					  block -> MetaPost(lex_stream, tex),
					  num_catches,
					  catches,
					  finally);
}

// --------------------------------------------------------------------------

MP_Ast *AstIntegerLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_IntLit(lex_stream.Line(LeftToken()),
						 lex_stream.Name(integer_literal_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstLongLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_IntLit(lex_stream.Line(LeftToken()),
						 lex_stream.Name(long_literal_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstFloatingPointLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_RealLit(lex_stream.Line(LeftToken()),
						  lex_stream.Name(floating_point_literal_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstDoubleLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_RealLit(lex_stream.Line(LeftToken()),
						  lex_stream.Name(double_literal_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstTrueLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_TrueLit(lex_stream.Line(LeftToken()));
}

// --------------------------------------------------------------------------

MP_Ast *AstFalseLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_FalseLit(lex_stream.Line(LeftToken()));
}

// --------------------------------------------------------------------------

MP_Ast *AstStringLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_StringLit(lex_stream.Line(LeftToken()),
							lex_stream.Name(string_literal_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstCharacterLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_StringLit(lex_stream.Line(LeftToken()),
							lex_stream.Name(character_literal_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstNullLiteral::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_NullLit(lex_stream.Line(LeftToken()));
}

// --------------------------------------------------------------------------

MP_Ast *AstThisExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_Identifier(lex_stream.Line(LeftToken()),
							 L"this",
							 Type() ? Type() -> Name() : NULL);
}

// --------------------------------------------------------------------------

MP_Ast *AstSuperExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_Identifier(lex_stream.Line(LeftToken()),
							 L"super",
							 Type() ? Type() -> Name() : NULL);
}

// --------------------------------------------------------------------------

MP_Ast *AstParenthesizedExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_Parens(lex_stream.Line(LeftToken()),
						 expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstTypeExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
    return type -> MetaPost(lex_stream, tex);
}

// --------------------------------------------------------------------------

MP_Ast *AstClassInstanceCreationExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_args = NumArguments();
	MP_Ast *args[num_args];

	//    if (dot_token_opt /* base_opt - see ast.h for explanation */)
	//		base_opt -> MetaPost(lex_stream, tex);
    
    for (int j = 0; j < num_args; j++)
		args[j] = this -> Argument(j) -> MetaPost(lex_stream, tex);
    if (class_body_opt)
    {
		/*		int save_method_num = method_num;
				method_num = 0;
				class_body_opt -> MetaPost(lex_stream, tex);
				method_num = save_method_num;*/
    }

	return new MP_New(lex_stream.Line(LeftToken()),
					  class_type -> MetaPost(lex_stream, tex),
					  num_args,
					  args);
}

// --------------------------------------------------------------------------

MP_Ast *AstDimExpr::MetaPost(LexStream& lex_stream, ostream& tex)
{
    return expression -> MetaPost(lex_stream, tex);
}

// --------------------------------------------------------------------------

MP_Ast *AstArrayCreationExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int num_dim_exprs = this -> NumDimExprs();
	MP_Ast* dim_exprs[num_dim_exprs];

    for (int j = 0; j < this -> NumDimExprs(); j++)
		dim_exprs[j] = DimExpr(j) -> MetaPost(lex_stream, tex);
	//    for (int k = 0; k < NumBrackets(); k++)
	//		cout << " []";
    return new MP_New(lex_stream.Line(LeftToken()),
					  array_type -> MetaPost(lex_stream, tex),
					  num_dim_exprs,
					  dim_exprs);
}

// --------------------------------------------------------------------------

MP_Ast *AstFieldAccess::MetaPost(LexStream& lex_stream, ostream& tex)
{
	TypeSymbol *s = Type();
	return new MP_Field(lex_stream.Line(LeftToken()),
						base -> MetaPost(lex_stream, tex),
						lex_stream.Name(identifier_token),
						Type() ? Type() -> Name() : NULL);
}

// --------------------------------------------------------------------------

static int count_subs(MethodSymbol *method_symbol)
{
	SymbolSet *s = method_symbol -> containing_type -> subtypes;
	TypeSymbol *t;
	int num_subs = 0;
	int i;
	for (t=(TypeSymbol*)s->FirstElement(); t; t=(TypeSymbol*)s->NextElement())
	{
		MethodSymbol *m = t->expanded_method_table->FindMethodShadowSymbol(method_symbol->name_symbol)->method_symbol;
		if (method_symbol -> NumFormalParameters() == m -> NumFormalParameters())
		{
			for (i = m -> NumFormalParameters() - 1; i >= 0; i--)
			{
				if (m -> FormalParameter(i) -> Type() != method_symbol -> FormalParameter(i) -> Type())
					break;
			}
			if (i < 0)
				num_subs += 1 + count_subs(m);
		}
	}
	return num_subs;
}

static void get_sub_names(MethodSymbol *method_symbol, wchar_t *subs[], int num_subs)
{
	SymbolSet *s = method_symbol -> containing_type -> subtypes;
	TypeSymbol *t;
	int i;
	for (t=(TypeSymbol*)s->FirstElement(); t; t=(TypeSymbol*)s->NextElement())
	{
		MethodSymbol *m = t->expanded_method_table->FindMethodShadowSymbol(method_symbol->name_symbol)->method_symbol;
		if (method_symbol -> NumFormalParameters() == m -> NumFormalParameters())
		{
			for (i = m -> NumFormalParameters() - 1; i >= 0; i--)
			{
				if (m -> FormalParameter(i) -> Type() != method_symbol -> FormalParameter(i) -> Type())
					break;
			}
			if (i < 0)
			{
				subs[num_subs] = m -> containing_type -> Name();
				get_sub_names(m, subs, num_subs + 1);
			}
		}
	}
}

MP_Ast *AstMethodInvocation::MetaPost(LexStream& lex_stream, ostream& tex)
{
	int i;
	MethodSymbol *method_symbol = (MethodSymbol *)symbol;
	int num_args = NumArguments();
	MP_Ast *args[num_args];
	wchar_t *argtypes[num_args];
	int num_throws = method_symbol -> NumThrows();
	wchar_t *throws[num_throws];
	MP_Throws *mp_throws = NULL;

	// build method argument arrays
	for (int i = 0; i < num_args; i++)
	{
		args[i] = this -> Argument(i) -> MetaPost(lex_stream, tex);
		argtypes[i] = this -> Argument(i) -> Type() -> Name();
	}

	// build throws data for method
	for (int i = 0; i < num_throws; i++)
		throws[i] = method_symbol -> Throws(i) -> Name();
	if (num_throws)
		mp_throws = new MP_Throws(lex_stream.Line(LeftToken()),
								  num_throws,
								  throws);

	// build array of overridden method names
	int num_subs = 1 + count_subs(method_symbol);
	wchar_t *subs[num_subs];
	subs[0] = method_symbol -> containing_type -> Name();
	get_sub_names(method_symbol, subs, 1);

	return new MP_ProcCall(lex_stream.Line(LeftToken()),
						   method -> MetaPost(lex_stream, tex),
						   num_args,
						   args,
						   argtypes,
						   Type() ? Type() -> Name() : NULL,
						   mp_throws,
						   num_subs,
						   subs);
}

// --------------------------------------------------------------------------

MP_Ast *AstArrayAccess::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_ArrayAccess(lex_stream.Line(LeftToken()),
							  base -> MetaPost(lex_stream, tex),
							  expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstPostUnaryExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
    return new MP_PostUnary(lex_stream.Line(LeftToken()),
							expression -> MetaPost(lex_stream, tex),
							lex_stream.Name(post_operator_token));
}

// --------------------------------------------------------------------------

MP_Ast *AstPreUnaryExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
    return new MP_PreUnary(lex_stream.Line(LeftToken()),
						   lex_stream.Name(pre_operator_token),
						   expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstCastExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	MP_Ast* type = NULL;

	if (type_opt)
		type = type_opt -> MetaPost(lex_stream, tex);
	return new MP_Cast(lex_stream.Line(LeftToken()),
					   type,
					   expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstBinaryExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	wchar_t *name = lex_stream.Name(binary_operator_token);

	if (name[0] == U_AMPERSAND and name[1] == U_AMPERSAND)
		return new MP_AndThen(lex_stream.Line(LeftToken()),
							  left_expression -> MetaPost(lex_stream, tex),
							  right_expression -> MetaPost(lex_stream, tex));
	else if (name[0] == U_BAR and name[1] == U_BAR)
		return new MP_OrThen(lex_stream.Line(LeftToken()),
							 left_expression -> MetaPost(lex_stream, tex),
							 right_expression -> MetaPost(lex_stream, tex));
	else
		return new MP_BinExpr(lex_stream.Line(LeftToken()),
							  left_expression -> MetaPost(lex_stream, tex),
							  new MP_Operator(lex_stream.Line(LeftToken()),
											  name),
							  right_expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstConditionalExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_Conditional(lex_stream.Line(LeftToken()),
							  test_expression -> MetaPost(lex_stream, tex),
							  true_expression -> MetaPost(lex_stream, tex),
							  false_expression -> MetaPost(lex_stream, tex));
}

// --------------------------------------------------------------------------

MP_Ast *AstAssignmentExpression::MetaPost(LexStream& lex_stream, ostream& tex)
{
	return new MP_BinExpr(lex_stream.Line(LeftToken()),
						  left_hand_side -> MetaPost(lex_stream, tex),
						  new MP_Operator(lex_stream.Line(LeftToken()),
										  L"="),
						  expression -> MetaPost(lex_stream, tex));
}
