
#include "mp_ast.h"
#include "mp_stacks.h"

int MP_Ast::num_boxes = 0;
bool MP_Ast::line_numbers;
static int method_num = 0;

static IntStack dummy_break_stack, case_break_stack, dummy_continue_stack;
static JumpStack break_stack, continue_stack;
static IntStack return_stack, throw_stack, num_throws_stack;
static ThrowCatchStack throw_catch_stack;
static IntStack loop_stack;
static CallStack call_stack;

// --------------------------------------------------------------------------

void MP_Ast::Box(ostream &out, char *box_type, 
				 int box1 = 0, int box2 = 0, int box3 = 0, int box4 = 0)
{
	StartBox(out, box_type);
	if (box1)
	{
		out << "(" << "b" << box1;
		if (box2)
		{
			out << ", b" << box2;
			if (box3)
			{
				out << ", b" << box3;
				if (box4)
					out << ", b" << box4;
			}
		}
		out << ")";
	}
	out << ";\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxArg(ostream &out, char *box_type, const char *arg,
					int box1 = 0, int box2 = 0, int box3 = 0, int box4 = 0)
{
	StartBox(out, box_type);
	out << "(";
	if (box1)
	{
		out << "b" << box1;
		if (box2)
		{
			out << ", b" << box2;
			if (box3)
			{
				out << ", b" << box3;
				if (box4)
					out << ", b" << box4;
			}
		}
		out << ", ";
	}
	out << arg << ");\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxBoxMath(ostream &out, char *box_type, int box, wchar_t *contents)
{
	StartBox(out, box_type);
	out << "(b" << box << ", btex $";
	Unicode::Out(out, contents);
	out << "$ etex);\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxDummy(ostream &out)
{
	StartBox(out, "boxit");
	out << "(btex \\Dummy etex);\n";
}
// --------------------------------------------------------------------------

void MP_Ast::BoxMacro(ostream &out, char *macro_name, const wchar_t *arg, 
					  const wchar_t *arg2 = NULL, const wchar_t *arg3 = NULL)
{
	StartBox(out, "boxit", true);
	out << "\\" << macro_name << "{";
	if (*arg == L'\\')
	{
		out << "\\";
		arg++;
	}
	TeXOut(out, arg);
	if (arg2)
	{
		out << "}{";
		if (*arg2 == L'\\')
		{
			out << "\\";
			arg2++;
		}
		TeXOut(out, arg2);
		if (arg3)
		{
			out << "}{";
			if (*arg3 == L'\\')
			{
				out << "\\";
				arg3++;
			}
			TeXOut(out, arg3);
		}
	}
	out << "} etex);\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxMacroText(ostream &out, char *macro_name, 
						  wchar_t *arg2 = NULL, wchar_t *arg3 = NULL)
{
	StartBox(out, "boxit", true);
	out << "\\" << macro_name << "{";
	PrintText(out);
	if (arg2)
	{
		out << "}{";
		TeXOut(out, arg2);
		if (arg3)
		{
			out << "}{";
			TeXOut(out, arg3);
		}
	}
	out << "} etex);\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxMath(ostream &out, char *box_type, wchar_t *contents)
{
	StartBox(out, box_type, true, true);
	Unicode::Out(out, contents);
	out << "$ etex);\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxMathBox(ostream &out, char *box_type, wchar_t *contents, int box)
{
	StartBox(out, box_type, true, true);
	Unicode::Out(out, contents);
	out << "$ etex)(b" << box << ");\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxString(ostream &out, char *box_type, wchar_t *string)
{
	StartBox(out, box_type);
	out << "(\"";
	Unicode::TeXOut(out, string);
	out << "\");\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxTeX(ostream &out, char *box_type, wchar_t *s1, 
					wchar_t *s2 = NULL, wchar_t *s3 = NULL)
{
	StartBox(out, box_type, true);
	TeXOut(out, s1);
	if (s2)
	{
		TeXOut(out, s2);
		if (s3)
			TeXOut(out, s3);
	}
	out << " etex);\n";
}

// --------------------------------------------------------------------------

void MP_Ast::BoxText(ostream& out, char *box_type)
{
	StartBox(out, box_type, true);
	PrintText(out);
	out << " etex);\n";
}

// --------------------------------------------------------------------------

void MP_Ast::LoopClose()
{
	while (dummy_break_stack.Size())
	{
		break_stack.Push(new jump(dummy_break_stack.Top(), num_boxes));
		dummy_break_stack.Pop();
	}
	while (dummy_continue_stack.Size())
	{
		continue_stack.Push(new jump(dummy_continue_stack.Top(), num_boxes));
		dummy_continue_stack.Pop();
	}
}

// --------------------------------------------------------------------------

void MP_Ast::StartBox(ostream &out, char *box_type, bool TeX, bool math)
{
	if (line_numbers)
		out << line_num << "\t";
	out << box_type << ".b" << ++num_boxes;
	if (TeX)
		out << "(btex ";
	if (math)
		out << "$";
}

// --------------------------------------------------------------------------

void MP_Ast::StatSeq(ostream &out, char *seq_string, MP_Ast *stats[], int num_stats)
{
	int box_nums[num_stats], i, j;

	for (i=0,j=0; i<num_stats; i++)
	{
		int prev_num_boxes = num_boxes;

		stats[i] -> Print(out);
		if (num_boxes > prev_num_boxes)
			box_nums[j++] = num_boxes;
	}
	if (j > 1)
	{
		StartBox(out, seq_string);
		out << "(";
		for (i=0; i<j; i++)
		{
			out << "b" << box_nums[i];
			if (i < j - 1)
				out << ", ";
		}
		out << ");\n";
	}
}

// --------------------------------------------------------------------------

void MP_Ast::TeXOut(ostream &out, const char *str)
{
	bool single_quote = false, double_quote = false;

	for (; *str; str++)
	{
		if (*str == '_' || *str == '{' || *str == '}' || *str == '#' ||
			/**str == '&' ||*/ *str == '$' || *str == '%')
			out << "\\";
		if (*str == '\\')
		{
			out << "{\\tt\\char92}";
			continue;
		}
		if (*str == '~')
		{
			out << "{\\char126}";
			continue;
		}
		if (*str == '^')
		{
			out << "{\\char70}";
			continue;
		}
		if (*str == ' ')
		{
			out << "{\\tt\\char`\\ }";
			continue;
		}
		if (*str == '\'')
		{
			if (single_quote)
			{
				out << "'";
				single_quote = false;
			}
			else
			{
				out << "`";
				single_quote = true;
			}
			continue;
		}
		if (*str == '"')
		{
			if (double_quote)
			{
				out << "''";
				double_quote = false;
			}
			else
			{
				out << "``";
				double_quote = true;
			}
			continue;
		}
		out << *str;
	}
}

// --------------------------------------------------------------------------

void MP_Ast::TeXOut(ostream &out, const wchar_t *str)
{
	bool single_quote = false, double_quote = false;

	for (; *str; str++)
	{
		if (*str == '_' || *str == '{' || *str == '}' || *str == '#' ||
			/**str == '&' ||*/ *str == '$' || *str == '%')
			out << "\\";
		if (*str == '\\')
		{
			out << "{\\tt\\char92}";
			continue;
		}
		if (*str == '~')
		{
			out << "{\\char126}";
			continue;
		}
		if (*str == '^')
		{
			out << "{\\char70}";
			continue;
		}
		if (*str == ' ')
		{
			out << "{\\tt\\char`\\ }";
			continue;
		}
		if (*str == '\'')
		{
			if (single_quote)
			{
				out << "'";
				single_quote = false;
			}
			else
			{
				out << "`";
				single_quote = true;
			}
			continue;
		}
		if (*str == '"')
		{
			if (double_quote)
			{
				out << "''";
				double_quote = false;
			}
			else
			{
				out << "``";
				double_quote = true;
			}
			continue;
		}
#ifdef EBCDIC
		out << (char) (*str > 0 && *str < 0x00C0 ? Code::ToEBCDIC(*str) : ' ');
#else
        out << (char) (*str > 0 && *str < 0x00C0 ? *str : ' ');
#endif
	}
}

// --------------------------------------------------------------------------

void MP_AndThen::Print(ostream& out)
{
	left -> Print(out);
	int l = num_boxes;
	right -> Print(out);
	Box(out, "AndThenExpr", l, num_boxes);
}

// --------------------------------------------------------------------------

void MP_ArrayAccess::Print(ostream& out)
{
	array -> Print(out);
	int a = num_boxes;
	expr -> Print(out);
	Box(out, "ArrayAccess", a, num_boxes);
}

// --------------------------------------------------------------------------

void MP_ArrayInit::Print(ostream& out)
{
	StartBox(out, "boxit", true);
	TeXOut(out, data);
	out << " etex);\n";
}

// --------------------------------------------------------------------------

void MP_BinExpr::Print(ostream& out)
{
	left -> Print(out);
	int l = num_boxes;
	op -> Print(out);
	int o = num_boxes;
	right -> Print(out);
	Box(out, "BinExpr", l, o, num_boxes);
}

// --------------------------------------------------------------------------

void MP_Break::Print(ostream& out)
{
	Box(out, "DummyBreakStat");
	if (loop_stack.Top())
		case_break_stack.Push(num_boxes);
	else
		dummy_break_stack.Push(num_boxes);
}

// --------------------------------------------------------------------------

void MP_Case::Print(ostream& out)
{
	int i, e, prev_b;

	expr -> Print(out);
	e = num_boxes;
	loop_stack.Push(1);
	for (i=0; i< num_cases; i++)
	{
		cases[i] -> prev_has_break = (i == 0) ? true : cases[i-1] -> has_break;
		cases[i] -> Print(out);
		if (i > 0)
			Box(out, "CaseBranchSeq", prev_b, num_boxes);
		prev_b = num_boxes;
	}
	if (default_case)
	{
		default_case -> Print(out);
		if (num_cases > 0)
			Box(out, "CaseBranchSeq", prev_b, num_boxes);
	}
	loop_stack.Pop();
	Box(out, "CaseStat", e, num_boxes);
	while (case_break_stack.Size())
	{
		break_stack.Push(new jump(case_break_stack.Top(), num_boxes));
		case_break_stack.Pop();
	}
}

// --------------------------------------------------------------------------

void MP_CaseBranch::Print(ostream& out)
{
	expr -> Print(out);
	int e = num_boxes;
	stat -> Print(out);
	if (prev_has_break)
		Box(out, "CaseBranch", e, num_boxes);
	else
		Box(out, "CaseFallThroughBranch", e, num_boxes);
}

// --------------------------------------------------------------------------

void MP_Cast::Print(ostream& out)
{
	if (type)
	{
		type -> Print(out);
		int t = num_boxes;
		expr -> Print(out);
		Box(out, "CastExpr", t, num_boxes);
	}
	else
	{
		//		out << "% implicit cast\n";
		expr -> Print(out);
	}
}

// --------------------------------------------------------------------------

void MP_Catch::Print(ostream& out)
{
	expr -> Print(out);
	int e = num_boxes;
	stat -> Print(out);
	Box(out, "CatchClause", e, num_boxes);
}

// --------------------------------------------------------------------------

void MP_Conditional::Print(ostream& out)
{
    expr -> Print(out);
	int e = num_boxes;
	true_stat -> Print(out);
	int t = num_boxes;
	false_stat -> Print(out);
	Box(out, "ConditionalExpr", e, t, num_boxes);
}

// --------------------------------------------------------------------------

void MP_Continue::Print(ostream& out)
{
	Box(out, "DummyContinueStat");
	dummy_continue_stack.Push(num_boxes);
}

// --------------------------------------------------------------------------

void MP_DefaultCaseBranch::Print(ostream& out)
{
	stat -> Print(out);
	if (prev_has_break)
		Box(out, "DefaultCaseBranch", num_boxes);
	else
		Box(out, "DefaultCaseFallThroughBranch", num_boxes);
}

// --------------------------------------------------------------------------

void MP_FalseLit::Print(ostream& out)
{
	BoxMacro(out, "Keyword", L"\\FALSE");
}

// --------------------------------------------------------------------------

void MP_Field::Print(ostream& out)
{
	base -> Print(out);
	int b = num_boxes;
	BoxMacro(out, "Identifier", field, type ? type : L"", L"");
	Box(out, "Field", b, num_boxes);
}

// --------------------------------------------------------------------------

void MP_Field::PrintText(ostream& out)
{
	Unicode::TeXOut(out, field);
}

// --------------------------------------------------------------------------

void MP_For::Print(ostream& out)
{
	if (!num_inits)
		BoxArg(out, "boxit", "");
	else
		StatSeq(out, "StatSeq", inits, num_inits);
	int i = num_boxes;
    if (condition)
		condition -> Print(out);
	else
		BoxArg(out, "boxit", "");
	int e = num_boxes;
	if (!num_updates)
		BoxArg(out, "boxit", "");
	else
		StatSeq(out, "StatSeq", updates, num_updates);
	int u = num_boxes;
	loop_stack.Push(0);
    stat -> Print(out);
	loop_stack.Pop();
	Box(out, "ForStat", i, e, u, num_boxes);
	LoopClose();
}

// --------------------------------------------------------------------------

void MP_Formal::Print(ostream& out)
{
	BoxMacroText(out, "Formal");
}

// --------------------------------------------------------------------------

void MP_Formal::PrintText(ostream& out)
{
	modifiers -> PrintText(out);
	variable_id -> PrintText(out);
}

// --------------------------------------------------------------------------

void MP_Formals::Print(ostream& out)
{
    for (int i = 0; i < num_formals; i++)
		formals[i] -> Print(out);
}

// --------------------------------------------------------------------------

void MP_Formals::PrintText(ostream& out)
{
    for (int i = 0; i < num_formals; i++)
    {
		formals[i] -> PrintText(out);
		if (i < num_formals - 1)
			out << ", ";
    }
}

// --------------------------------------------------------------------------

void MP_Identifier::Print(ostream &out)
{
	BoxMacro(out, "Identifier", name, type ? type : L"", owner ? owner : L"");
}

// --------------------------------------------------------------------------

void MP_If::Print(ostream& out)
{
	int i, e, prev_b;

	exprs[0] -> Print(out);
	e = num_boxes;
	stats[0] -> Print(out);
	Box(out, "IfThenBranch", e, num_boxes);
	prev_b = num_boxes;
	for (i=1; i< num_ifs; i++)
	{
		exprs[i] -> Print(out);
		e = num_boxes;
		stats[i] -> Print(out);
		Box(out, "ElseIfThenBranch", e, num_boxes);
		if (i > 0)
			Box(out, "IfThenBranchSeq", prev_b, num_boxes);
		prev_b = num_boxes;
	}
	if (else_stat)
	{
		else_stat -> Print(out);
		Box(out, "ElseBranch", num_boxes);
		Box(out, "IfThenBranchSeq", prev_b, num_boxes);
	}
	Box(out, "IfThenStat", num_boxes);
}

// --------------------------------------------------------------------------

void MP_IntLit::Print(ostream& out)
{
	BoxMacro(out, "Number", value);
}

// --------------------------------------------------------------------------

void MP_Modifiers::Print(ostream& out)
{
}

// --------------------------------------------------------------------------

void MP_Modifiers::PrintText(ostream& out)
{
    for (int i = 0; i < num_modifiers; i++)
    {
		Unicode::TeXOut(out, modifiers[i]);
		out << " ";
    }
}

// --------------------------------------------------------------------------

void MP_New::Print(ostream& out)
{
    type -> Print(out);
	int t = num_boxes;
	if (num_args)
		StatSeq(out, "ArgSeq", args, num_args);
	else
		BoxDummy(out);
	Box(out, "NewExpr", t, num_boxes);
}

// --------------------------------------------------------------------------

void MP_NullLit::Print(ostream& out)
{
	BoxMacro(out, "Keyword", L"\\NULL");
}

// --------------------------------------------------------------------------

void MP_Operator::Print(ostream& out)
{
	if (*op == L'*')
		BoxMacro(out, "Operator", L"\\times");
	else if (*op == L'%')
		BoxMacro(out, "Operator", L"\\bmod");
	else if (*op == L'&')
		BoxMacro(out, "Operator", L"\\&");
	else if (*op == L'^')
		BoxMacro(out, "Operator", L"\\hbox{\\char70}");
	else if (*op == L'!')
		BoxMacro(out, "Operator", L"\\ne");
	else if (*op == L'<' && *(op + 1) == L'=')
		BoxMacro(out, "Operator", L"\\le");
	else if (*op == L'>' && *(op + 1) == L'=')
		BoxMacro(out, "Operator", L"\\ge");
	else if (*op == L'=')
		if (*(op + 1) == L'=')
			BoxMacro(out, "Operator", L"\\equiv");
		else
			BoxMacro(out, "Operator", L"\\leftarrow");
	else if (*op == L'i')
		BoxMacro(out, "OperatorText", L"\\INSTANCEOF");
	else
		BoxMacro(out, "Operator", op);
}

// --------------------------------------------------------------------------

void MP_OrThen::Print(ostream& out)
{
	left -> Print(out);
	int l = num_boxes;
	right -> Print(out);
	Box(out, "OrThenExpr", l, num_boxes);
}

// --------------------------------------------------------------------------

void MP_Parens::Print(ostream& out)
{
    expr -> Print(out);
	Box(out, "Parens", num_boxes);
}

// --------------------------------------------------------------------------

void MP_PostUnary::Print(ostream& out)
{
    expr -> Print(out);
	int e = num_boxes;
	StartBox(out, "boxit", true, true);
	if (*op == L'~')
		out << "\\hbox{\\char126}";
	else
		Unicode::Out(out, op);
	out << "$etex);\n";
	Box(out, "PostUnaryExpr", e, num_boxes);
}

// --------------------------------------------------------------------------

void MP_PreUnary::Print(ostream& out)
{
	StartBox(out, "boxit", true, true);
	if (*op == L'~')
		out << "\\hbox{\\char126}";
	else
		Unicode::Out(out, op);
	out << "$etex);\n";
	int o = num_boxes;
    expr -> Print(out);
	Box(out, "PreUnaryExpr", o, num_boxes);
}

// --------------------------------------------------------------------------

void MP_ProcCall::Print(ostream& out)
{
	proc -> Print(out);
	int p = num_boxes;

	StartBox(out, "boxit", true);
	out << "\\ProcCallSignature{";
	if (return_type)
		Unicode::Out(out, return_type);
	out << "}{";
	for (int i=0; i<num_args; i++)
	{
		if (i)
			out << ", ";
		Unicode::Out(out, argtypes[i]);
	}
	out << "} etex);\n";

	if (0 == num_args)
		BoxArg(out, "boxit", "btex \\Dummy etex");
	else if (1 == num_args)
		args[0] -> Print(out);
	else
		StatSeq(out, "ArgSeq", args, num_args);
	int a = num_boxes;

	if (throws)
		throws -> Print(out);
	int t = num_boxes;

	if (1 == num_subs)
	{
		BoxArg(out, "boxit", "btex \\Dummy etex");
	}
	else
	{
		int s[num_subs];

		for (int i=0; i<num_subs; i++)
		{
			out << "boxit.b" << ++num_boxes << "(btex \\SubMethod{";
			Unicode::TeXOut(out, subs[i]);
			out << ".";
			proc -> PrintText(out);
			out << "} etex);\n";
			s[i] = num_boxes;
		}
		out << "SubclassSeq.b" << ++num_boxes << "(";
		for (int i=0; i<num_subs; i++)
		{
			if (i)
				out << ", ";
			out << "b" << s[i];
		}
		out << ");\n";
	}

	if (throws)
		StartBox(out, "ProcCallThrowsStat");
	else
		StartBox(out, "ProcCallStat");
	out << "(b" << p << ", b" << p + 1 << ", b" << a;
	if (throws)
		out << ", b" << t;
	out << ");\n";

	call_stack.Push(new call(num_boxes, num_boxes - 1));
}

// --------------------------------------------------------------------------

void MP_Procedure::Print(ostream& out)
{
	num_boxes = 0;
	out << "beginfig(" << ++method_num << ");\n";

	if (stat)
		stat -> Print(out);
	if (!num_boxes)
		BoxArg(out, "boxit", "");

	out << "Procedure(btex ";
	modifiers -> PrintText(out);
	if (type)
	{
		type -> PrintText(out);
		out << " ";
	}
	Unicode::TeXOut(out, name);
	out << "(";
	formals -> PrintText(out);
	out << ")";
	if (!num_boxes)
		out << ";";
	out << " etex)(b" << num_boxes << ");\n";

	while (break_stack.Size())
	{
		out << "BreakArrow(b" << break_stack.Top() -> loop;
		out << ", b" << break_stack.Top() -> statement << ");\n";
		delete break_stack.Top();
		break_stack.Pop();
	}
	while (continue_stack.Size())
	{
		out << "ContinueArrow(b" << continue_stack.Top() -> loop;
		out << ", b" << continue_stack.Top() -> statement << ");\n";
		delete continue_stack.Top();
		continue_stack.Pop();
	}
	while (return_stack.Size())
	{
		out << "ReturnArrow(b" << return_stack.Top() << ");\n";
		return_stack.Pop();
	}
	while (throw_stack.Size())
	{
		int box_num = throw_stack.Top();
		out << "ThrowStat(b" << box_num << ", b" << box_num - 1 << ");\n";
		throw_stack.Pop();
	}
	while (throw_catch_stack.Size())
	{
		int b = throw_catch_stack.Top() -> statement;
		int c = throw_catch_stack.Top() -> clause;
		out << "ThrowCatchStat(b" << b << ", b" << c << ", b" << b - 1 << ");\n";
		delete throw_catch_stack.Top();
		throw_catch_stack.Pop();
	}
	while (call_stack.Size())
	{
		out << "CallArrow(b" << call_stack.Top()->statement;
		out << ", b" << call_stack.Top()->subs << ");\n";
		delete call_stack.Top();
		call_stack.Pop();
	}
	out << "endfig;\n\n";
}

// --------------------------------------------------------------------------

void MP_RealLit::Print(ostream& out)
{
	BoxMacro(out, "Number", value);
}

// --------------------------------------------------------------------------

void MP_Repeat::Print(ostream& out)
{
	loop_stack.Push(0);
    stat -> Print(out);
	loop_stack.Pop();
	int s = num_boxes;
	expr -> Print(out);
	Box(out, "RepeatStat", s, num_boxes);
	LoopClose();
}

// --------------------------------------------------------------------------

void MP_Return::Print(ostream& out)
{
	if (expr)
	{
		expr -> Print(out);
		Box(out, "ReturnStat", num_boxes);
	}
	else
		Box(out, "DummyReturnStat");
	return_stack.Push(num_boxes);
}

// --------------------------------------------------------------------------

void MP_StatSeq::Print(ostream& out)
{
	StatSeq(out, "StatSeq", stats, num_stats);
	if (0 == num_stats)
		BoxArg(out, "boxit", "");
}

// --------------------------------------------------------------------------

void MP_StringLit::Print(ostream& out)
{
	BoxMacro(out, "String", value);
}

// --------------------------------------------------------------------------

void MP_Throw::Print(ostream& out)
{
	expr -> Print(out);
	Box(out, "DummyThrowStat");
	throw_stack.Push(num_boxes);
	if (num_throws_stack.Size())
	{
		int num_throws = num_throws_stack.Top();
		num_throws_stack.Pop();
		num_throws_stack.Push(num_throws + 1);
	}
}

// --------------------------------------------------------------------------

void MP_Throws::Print(ostream& out)
{
	BoxMacroText(out, "ProcThrows");
}

// --------------------------------------------------------------------------

void MP_Throws::PrintText(ostream& out)
{
    for (int i = 0; i < num_throws; i++)
    {
		Unicode::TeXOut(out, throws[i]);
		if (i < num_throws - 1)
			out << ", ";
    }
}

// --------------------------------------------------------------------------

void MP_TrueLit::Print(ostream& out)
{
	BoxMacro(out, "Keyword", L"\\TRUE");
}

// --------------------------------------------------------------------------

void MP_Try::Print(ostream& out)
{
	int i;

	num_throws_stack.Push(0);
	stat -> Print(out);
	int num_throws = num_throws_stack.Top();
	num_throws_stack.Pop();
	num_throws_stack.Push(0);

	int b = num_boxes;
	if (num_catches)
		StatSeq(out, "CatchClauseSeq", (MP_Ast **)catches, num_catches);
	else
		BoxArg(out, "boxit", "");
	int c = num_boxes;
    if (finally)
		finally -> Print(out);
	else
		BoxArg(out, "boxit", "");
	Box(out, "TryBlock", b, c, num_boxes);

	int num_skip = num_throws_stack.Top();
	num_throws_stack.Pop();
	IntStack skip_stack;
	for (i = 0; i < num_skip; i++)
	{
		skip_stack.Push(throw_stack.Top());
		throw_stack.Pop();
	}
	for (i = 0; i < num_throws; i++)
	{
		throw_catch_stack.Push(new ThrowCatch(throw_stack.Top(), c));
		throw_stack.Pop();
	}
	for (i = 0; i < num_skip; i++)
	{
		throw_stack.Push(skip_stack.Top());
		skip_stack.Pop();
	}
}

// --------------------------------------------------------------------------

void MP_Type::Print(ostream &out)
{
	BoxMacroText(out, "Type");
}

// --------------------------------------------------------------------------

void MP_Type::PrintText(ostream &out)
{
	Unicode::TeXOut(out, type);
    for (int i = 0; i < num_dimens; i++)
		out << "[]";
}

// --------------------------------------------------------------------------

void MP_VariableDeclaration::Print(ostream& out)
{
	int box_nums[num_decls], i;
	
	for (i=0; i<num_decls; i++)
	{
		decls[i] -> Print(out);
		box_nums[i] = num_boxes;
	}
	StartBox(out, "VariableDeclaration", true);
	out << "\\Type{";
	type -> PrintText(out);
	out << "} etex)(";
	for (i=0; i<num_decls; i++)
	{
		out << "b" << box_nums[i];
		if (i < num_decls - 1)
			out << ", ";
	}
	out << ");\n";
}

// --------------------------------------------------------------------------

void MP_VariableDeclaration::PrintText(ostream& out)
{
	type -> PrintText(out);
	out << " ";
	for (int i=0; i<num_decls; i++)
	{
		decls[i] -> PrintText(out);
		if (i < num_decls - 1)
			out << ", ";
	}
}

// --------------------------------------------------------------------------

void MP_VariableDeclarationId::Print(ostream &out)
{
	BoxMacroText(out, "Identifier", L"", L"");
}

// --------------------------------------------------------------------------

void MP_VariableDeclarationId::PrintText(ostream &out)
{
	Unicode::TeXOut(out, name);
    for (int i = 0; i < num_dimens; i++)
		out << "[]";
}

// --------------------------------------------------------------------------

void MP_While::Print(ostream& out)
{
	expr -> Print(out);
	int e = num_boxes;
	if (stat)
	{
		loop_stack.Push(0);
		stat -> Print(out);
		loop_stack.Pop();
	}
	else
		BoxArg(out, "boxit", "");
	Box(out, "WhileStat", e, num_boxes);
	LoopClose();
}

